/**
*	WARNING: Do Not Remove This Section
*
*	$RCSfile: simulation_utils.h,v $
*	$Revision: 1.20 $
*	$Date: 2008/06/02 16:02:57 $
*	$Author: i.zapreev $
*
*	MRMC is a model checker for discrete-time and continuous-time Markov reward models.
*	It supports reward extensions of PCTL and CSL (PRCTL and CSRL), and allows for the
*	automated verification of properties concerning long-run and instantaneous rewards
*	as well as cumulative rewards.
*
*	Copyright (C) The University of Twente, 2004-2007.
*	Authors: Ivan Zapreev
*
*	This program is free software; you can redistribute it and/or
*	modify it under the terms of the GNU General Public License
*	as published by the Free Software Foundation; either version 2
*	of the License, or (at your option) any later version.
*
*	This program is distributed in the hope that it will be useful,
*	but WITHOUT ANY WARRANTY; without even the implied warranty of
*	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*	GNU General Public License for more details.
*
*	You should have received a copy of the GNU General Public License
*	along with this program; if not, write to the Free Software
*	Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*
*	Our contacts:
*		Formal Methods and Tools Group, University of Twente,
*		P.O. Box 217, 7500 AE Enschede, The Netherlands,
*		Phone: +31 53 4893767, Fax: +31 53 4893247,
*		E-mail: mrmc@cs.utwente.nl
*
*	Source description: This is a header file for the simulation engine
*	here we intend to define the utility functions and data structures.
*/

#include "macro.h"

#include "sample_vec.h"
#include "sparse.h"
#include "bitset.h"

#ifndef SIMULATION_UTILS_H

#define SIMULATION_UTILS_H
	
	/****************************************************************************/
	/**************CHECK THE CONF. INT. AGAINST THE PROB. CONSTRAINT*************/
	/****************************************************************************/

	/**
	* This method sorts out the pAlwaysBitSet and pNeverBitSet states and updates the sets 
	* pYesBitSet and pNoBitSet with respect to the probability constraint
	* "comparator prob_bound".
	* NOTE: both pAlwaysBitSet and pNeverBitSet are allowed to be NULL
	* @param isSimOneInitState TRUE if we simulate one initial state only
	* @param pYesBitSet this bitset is going to be filled with the states which
	*			satisfy the formula
	* @param pNoBitSet this bitset is going to be filled with the states which
	*			do not satisfy the formula
	* @param pAlwaysBitSet the states that satisfy the formula with probability >= 1.0
	* @param pNeverBitSet the states that satisfy the formula with probability <= 0.0
	* @param initial_state the initial state index if isSimOneInitState == TRUE
	* @param comparator the comparator, one of: 
	*	COMPARATOR_SF_GREATER, COMPARATOR_SF_GREATER_OR_EQUAL,
	*	COMPARATOR_SF_LESS, COMPARATOR_SF_LESS_OR_EQUAL
	* @param prob_bound the probability bound
	*/
	extern inline void considerAlwaysAndNeverStates( const BOOL isSimOneInitState,
						bitset * pYesBitSet, bitset * pNoBitSet,
						const bitset * pAlwaysBitSet, const bitset * pNeverBitSet,
						const int initial_state, const int comparator,
						const double prob_bound );
	
	/**
	* This function updated the pYesBitSet and pNoBitSet bitsets with respect
	* to the model checking result mc_result for the initial_state state.
	* @param mc_result the model checking result in the three valued logic
	* @param initial_state the initial state for which model checking was done
	* @param pYesBitSet the initial_state'th bit of this set will be set on
	* 			if mc_result == TVL_TT
	* @param pNoBitSet the initial_state'th bit of this set will be set on
	* 			if mc_result == TVL_FF
	*/
	extern inline void markYesNoSetEntree(const TV_LOGIC mc_result, const int initial_state,
					bitset * pYesBitSet, bitset * pNoBitSet);
	
	/**
	* This method checks if "probability comparator prob_bound" holds.
	* @param probability the left side of the constraint, i.e. the probability
	* @param comparator the comparator, one of: 
	*	COMPARATOR_SF_GREATER, COMPARATOR_SF_GREATER_OR_EQUAL,
	*	COMPARATOR_SF_LESS, COMPARATOR_SF_LESS_OR_EQUAL
	* @param prob_bound the probability bound
	* @return TRUE if "probability comparator prob_bound" holds, otherwise FALSE
	*/
	extern inline BOOL doesComparisonHold( const double probability, const int comparator, const double prob_bound );
	
	/**
	* This is an implementation of the checkBoundVSConfInt function from the PhD thesis
	* of Ivan S. Zapreev, see Chapter 6.
	* Plus if "ciRightBorder - ciLeftBorder >= indiff_width" then the method returns TVL_NN
	* @param comparator the comparator, one of: 
	*	COMPARATOR_SF_GREATER, COMPARATOR_SF_GREATER_OR_EQUAL,
	*	COMPARATOR_SF_LESS, COMPARATOR_SF_LESS_OR_EQUAL
	* @param prob_bound the probability bound
	* @param ciLeftBorder the left border of the conf. int.
	* @param ciRightBorder the right border of the conf. int.
	* @param indiff_width the indifference-region width
	* @return returns one of: TVL_TT, TVL_FF, TVL_NN
	*/
	TV_LOGIC checkBoundVSConfInt( const int comparator, const double prob_bound, const double ciLeftBorder,
					const double ciRightBorder, const double indiff_width );

	/**
	* This function checks the precondition of the conf. int. like it is done in checkBoundVSConfInt(...)
	* Namely we check if the width of the confidence interval ( ciRightBorder - ciLeftBorder ) for the state
	* initial_state is >= indiff_width. Then the sample size is insufficient and we have an error state which
	* we indicate by setting isIndiffErrStates to TRUE and adding the error state to the pErrStatesBitset bitset.
	* 
	* NOTE: Just in case we also check if the conf. int. is invalid i.e. ( ciRightBorder - ciLeftBorder ) < 0
	* 
	* @param ciLeftBorder the left border of the conf int.
	* @param ciRightBorder the right border of the conf. int.
	* @param initial_state the initial state for which the conf. int. was computed
	* @param indiff_width the supremum of the allowed conf. int. width
	* @param pIsIndiffErrStates the pointer to the boolean variable that should be TRUE if an error state exists
	* @param pErrStatesBitset the bitset containing the indexes of error states
	*/
	inline void checkErrorConfIntState(const double ciLeftBorder, const double ciRightBorder, const int initial_state,
						const double indiff_width, BOOL * pIsIndiffErrStates, const bitset * pErrStatesBitset);
		
	/****************************************************************************/
	/**************THE INCREASE FUNCTION WITH THE STEP AND MAX BOUND*************/
	/****************************************************************************/
	
	/**
	* This method does the following:
	* 1. (* pTarget) += inc_step;
	* 2. if (* pTarget) > max_val then
	*	(* pTarget) = max_val;
	* @param pTarget the pointer to the variable that will be increased.
	* @param inc_step the increment step
	* @param max_val the maximum value
	*/
	extern inline void increment( int * pTarget, const int inc_step, const int max_val );
	
	/****************************************************************************/
	/**************THE CONF. INT. COMPUTING FUNCTIONS FOR DIFF SAMPLES***********/
	/****************************************************************************/
	
	/* The standard conf int as resulted from the central limit theorem */
	#define STANDARD_CONF_INT 0
	/* The improved version of the standard conf. int. for Bernoulli Trials */
	#define AGRESTI_COULL_CONF_INT 1
	
	/**
	* This method computes different conf intervals based on the provided values
	* Note: this method allows to use two different parameters for the left and
	* the right conf. int. borders. This is needed for computing the intervals
	* in case of the unbounded-reachability problem.
	* @param conf_zeta the zeta value derived from the confidence
	* @param pCiLeftBorder the pointer to the left conf. int. border
	* @param pCiRightBorder the pointer to the right conf. int. border
	* @param left_param the parameter for computing the left conf. int. border
	* @param right_param the parameter for computing the right conf. int. border
	* @param theConfIntType indicate the kind of the conf. int. to be computed:
	*				STANDARD_CONF_INT, AGRESTI_COULL_CONF_INT, etc.
	*/
	extern inline void computeBordersConfInt(const double conf_zeta, double * pCiLeftBorder,
					double * pCiRightBorder, const double sample_size,
					const double left_param, const double right_param,
					const int theConfIntType);
	
	/**
	* This method computew the conf. int. borders. for the given sample, i.e.
	* we compute and set: (*pCiLeftBorder), (*pCiRightBorder).
	* NOTE: The borders that do not fit into the interval R[0,1] are truncated.
	* @param conf_zeta the zeta value derived from the confidence
	* @param pCiLeftBorder the pointer to the left conf. int. border
	* @param pCiRightBorder the pointer to the right conf. int. border
	* @param pSampleVecUntil the sample we use to derive the conf int
	* @param theConfIntType indicate the kind of the conf. int. to be computed:
	*				STANDARD_CONF_INT, AGRESTI_COULL_CONF_INT, etc.
	*/
	extern inline void computeBordersUU( const double conf_zeta, double * pCiLeftBorder,
					double * pCiRightBorder, const PTSampleVecUntil pSampleVecUntil,
					const int theConfIntType );
					
	/**
	* This method computes the conf. int. borders of steady-state simulation for 
	* the given sample, i.e. we compute and set: (*pCiLeftBorder), (*pCiRightBorder).
	* For more information see PhD Thesis of Ivan S. Zapreev.
	* NOTE: The borders that do not fit into the interval R[0,1] are truncated.
	* @param reg_cycles the number of samples, i.e. regeneration cycles calculated so far
	* @param conf_zeta the zeta value derived from the confidence
	* @param pCiLeftBorder the pointer to the left conf. int. border
	* @param pCiRightBorder the pointer to the right conf. int. border
	* @param pSampleVecSteady the sample we use to derive the conf int
	*/
	extern inline void computeBordersConfIntSS( const int reg_cycles, const double conf_zeta, double * pCiLeftBorder, 
		double * pCiRightBorder, const PTSampleVecSteady pSampleVecSteady );
		
	/****************************************************************************/
	/**********THE FUNCTION THAT SIMULATES THE SAMPLE FOR UNBOUNDED UNTIL********/
	/****************************************************************************/
	
	/**
	* This function is an implementation of the extendSamples function from the PhD Thesis
	* of Ivan S. Zapreev. It either add new observations to the sample and simulates them
	* or it increases the simulation depth of every observation in the sample.
	* NOTE: This method can increase the sample size and the simulation depth at the same time.
	* WARNING: We simulate the states ASSUMING there are no self loops!
	* @param pStateSpace the sparse matrix of the embedded DTMC with good ad bad
	*			states made absorbing
	* @param pSampleVecUntil the sample vector obtained on the previous iteration
	* @param new_sample_size the new sample size
	* @param new_simulation_depth the new simulation depth
	* @param pGoodStates the bitsets containing all the good absorbing states
	* @param pTransientStates the bitsets containing all the transient states
	*/
	extern void simulateSampleVectorUnbUntilDTMC( const sparse * pStateSpace, PTSampleVecUntil pSampleVecUntil,
							const int new_sample_size, const int new_simulation_depth,
							const bitset * pGoodStates, const bitset * pTransientStates );
	
#endif
