/**
*	WARNING: Do Not Remove This Section
*
*	$RCSfile: simulation_utils.c,v $
*	$Revision: 1.37 $
*	$Date: 2008/06/03 08:48:59 $
*	$Author: i.zapreev $
*
*	MRMC is a model checker for discrete-time and continuous-time Markov reward models.
*	It supports reward extensions of PCTL and CSL (PRCTL and CSRL), and allows for the
*	automated verification of properties concerning long-run and instantaneous rewards
*	as well as cumulative rewards.
*
*	Copyright (C) The University of Twente, 2004-2007.
*	Authors: Ivan Zapreev
*
*	This program is free software; you can redistribute it and/or
*	modify it under the terms of the GNU General Public License
*	as published by the Free Software Foundation; either version 2
*	of the License, or (at your option) any later version.
*
*	This program is distributed in the hope that it will be useful,
*	but WITHOUT ANY WARRANTY; without even the implied warranty of
*	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*	GNU General Public License for more details.
*
*	You should have received a copy of the GNU General Public License
*	along with this program; if not, write to the Free Software
*	Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*
*	Our contacts:
*		Formal Methods and Tools Group, University of Twente,
*		P.O. Box 217, 7500 AE Enschede, The Netherlands,
*		Phone: +31 53 4893767, Fax: +31 53 4893247,
*		E-mail: mrmc@cs.utwente.nl
*
*	Source description: This is a header file for the simulation engine
*	here we intend to define the utility functions and data structures.
*/

#include "simulation_utils.h"

#include "parser_to_tree.h"
#include "rand_num_generator.h"

/****************************************************************************/
/**************CHECK THE CONF. INT. AGAINST THE PROB. CONSTRAINT*************/
/****************************************************************************/

/**
* This method sorts out the pAlwaysBitSet and pNeverBitSet states and updates the sets 
* pYesBitSet and pNoBitSet with respect to the probability constraint
* "comparator prob_bound".
* NOTE: both pAlwaysBitSet and pNeverBitSet are allowed to be NULL
* @param isSimOneInitState TRUE if we simulate one initial state only
* @param pYesBitSet this bitset is going to be filled with the states which
*			satisfy the formula
* @param pNoBitSet this bitset is going to be filled with the states which
*			do not satisfy the formula
* @param pAlwaysBitSet the states that satisfy the formula with probability >= 1.0
* @param pNeverBitSet the states that satisfy the formula with probability <= 0.0
* @param initial_state the initial state index if isSimOneInitState == TRUE
* @param comparator the comparator, one of: 
*	COMPARATOR_SF_GREATER, COMPARATOR_SF_GREATER_OR_EQUAL,
*	COMPARATOR_SF_LESS, COMPARATOR_SF_LESS_OR_EQUAL
* @param prob_bound the probability bound
*/
inline void considerAlwaysAndNeverStates( const BOOL isSimOneInitState,
					bitset * pYesBitSet, bitset * pNoBitSet,
					const bitset * pAlwaysBitSet, const bitset * pNeverBitSet,
					const int initial_state, const int comparator,
					const double prob_bound ){
	/* Check if the probability constraint is satisfied for probabilities 1.0 or 0.0 */
	BOOL isAlwaysSetGood = doesComparisonHold( 1.0, comparator, prob_bound );
	BOOL isNeverSetGood =  doesComparisonHold( 0.0, comparator, prob_bound );
	
	if( isSimOneInitState ){
		/* If we simulate just one initial state than all the other data is not iteresting */
		if( ( pAlwaysBitSet != NULL ) && get_bit_val( pAlwaysBitSet, initial_state ) ){
			if( isAlwaysSetGood ){
				set_bit_val( pYesBitSet, initial_state, BIT_ON );
			} else {
				set_bit_val( pNoBitSet, initial_state, BIT_ON );
			}
		} else {
			if( ( pNeverBitSet != NULL ) && get_bit_val( pNeverBitSet, initial_state ) ){
				if( isNeverSetGood ){
					set_bit_val( pYesBitSet, initial_state, BIT_ON );
				} else {
					set_bit_val( pNoBitSet, initial_state, BIT_ON );
				}
			}
		}
	} else {
		/* All pAlwaysBitSet states satisfy the formula with probability >= 1.0 */
		if( pAlwaysBitSet != NULL ){
			if( isAlwaysSetGood ){
				or_result( pAlwaysBitSet, pYesBitSet );
			} else {
				or_result( pAlwaysBitSet, pNoBitSet );
			}
		}
		/* All pNeverBitSet states satisfy the formula with probability <= 0.0 */
		if( pNeverBitSet != NULL ){
			if( isNeverSetGood ){
				or_result( pNeverBitSet, pYesBitSet );
			} else {
				or_result( pNeverBitSet, pNoBitSet );
			}
		}
	}
}

/**
 * This function updated the pYesBitSet and pNoBitSet bitsets with respect
 * to the model checking result mc_result for the initial_state state.
 * @param mc_result the model checking result in the three valued logic
 * @param initial_state the initial state for which model checking was done
 * @param pYesBitSet the initial_state'th bit of this set will be set on
 * 			if mc_result == TVL_TT
 * @param pNoBitSet the initial_state'th bit of this set will be set on
 * 			if mc_result == TVL_FF
 */
inline void markYesNoSetEntree(const TV_LOGIC mc_result, const int initial_state,
				bitset * pYesBitSet, bitset * pNoBitSet){
	switch( mc_result ){
		case TVL_TT:
			set_bit_val( pYesBitSet, initial_state, BIT_ON);
			break;
		case TVL_FF:
			set_bit_val( pNoBitSet, initial_state, BIT_ON);
			break;
	}
}

/**
* This method checks if "probability comparator prob_bound" holds.
* @param probability the left side of the constraint, i.e. the probability
* @param comparator the comparator, one of: 
*	COMPARATOR_SF_GREATER, COMPARATOR_SF_GREATER_OR_EQUAL,
*	COMPARATOR_SF_LESS, COMPARATOR_SF_LESS_OR_EQUAL
* @param prob_bound the probability bound
* @return TRUE if "probability comparator prob_bound" holds, otherwise FALSE
*/
inline BOOL doesComparisonHold( const double probability, const int comparator, const double prob_bound ){
	switch( comparator ){
		case COMPARATOR_SF_GREATER:
			return ( probability > prob_bound ? TRUE : FALSE );
		case COMPARATOR_SF_GREATER_OR_EQUAL:
			return ( probability >= prob_bound ? TRUE : FALSE );
		case COMPARATOR_SF_LESS:
			return ( probability < prob_bound ? TRUE : FALSE );
		case COMPARATOR_SF_LESS_OR_EQUAL:
			return ( probability <= prob_bound ? TRUE : FALSE );
		default:
			printf("ERROR: An unknown comparator type %d.\n", comparator);
			exit(1);
	}
}

static inline TV_LOGIC checkBoundVSConfIntHelper(BOOL theTRUECond, BOOL theFALSECond){
	TV_LOGIC tvl_result = TVL_NN;
	if( theTRUECond ){
		tvl_result = TVL_TT;
	} else {
		if( theFALSECond ){
			tvl_result = TVL_FF;
		}
	}
	return tvl_result;
}

/**
* This is an implementation of the checkBoundVSConfInt function from the PhD thesis
* of Ivan S. Zapreev, see Chapter 6.
* Plus if "ciRightBorder - ciLeftBorder >= indiff_width" then the method returns TVL_NN
* @param comparator the comparator, one of: 
*	COMPARATOR_SF_GREATER, COMPARATOR_SF_GREATER_OR_EQUAL,
*	COMPARATOR_SF_LESS, COMPARATOR_SF_LESS_OR_EQUAL
* @param prob_bound the probability bound
* @param ciLeftBorder the left border of the conf. int.
* @param ciRightBorder the right border of the conf. int.
* @param indiff_width the indifference-region width
* @return returns one of: TVL_TT, TVL_FF, TVL_NN
*/
TV_LOGIC checkBoundVSConfInt( const int comparator, const double prob_bound, const double ciLeftBorder,
				const double ciRightBorder, const double indiff_width ){
	TV_LOGIC tvl_result = TVL_NN;
	
	/* First check if it is a valid conf. int. */
	if( ( ciRightBorder - ciLeftBorder ) < indiff_width ){
		/* Distinguish the cases based on the comparator type */
		switch( comparator ){
			case COMPARATOR_SF_LESS_OR_EQUAL:
				tvl_result = checkBoundVSConfIntHelper( prob_bound >= ciRightBorder,
									prob_bound < ciLeftBorder );
				break;
			case COMPARATOR_SF_LESS:
				tvl_result = checkBoundVSConfIntHelper( prob_bound > ciRightBorder,
									prob_bound <= ciLeftBorder );
				break;
			case COMPARATOR_SF_GREATER_OR_EQUAL:
				tvl_result = checkBoundVSConfIntHelper( prob_bound <= ciLeftBorder,
									prob_bound > ciRightBorder );
				break;
			case COMPARATOR_SF_GREATER:
				tvl_result = checkBoundVSConfIntHelper( prob_bound < ciLeftBorder,
									prob_bound >= ciRightBorder );
				break;
			default:
				printf("ERROR: An unexpected comparator type: %d.\n", comparator);
				exit(1);
		}
	}
	
	return tvl_result;
}

/**
* This function checks the precondition of the conf. int. like it is done in checkBoundVSConfInt(...)
* Namely we check if the width of the confidence interval ( ciRightBorder - ciLeftBorder ) for the state
* initial_state is >= indiff_width. Then the sample size is insufficient and we have an error state which
* we indicate by setting isIndiffErrStates to TRUE and adding the error state to the pErrStatesBitset bitset.
* 
* NOTE: Just in case we also check if the conf. int. is invalid i.e. ( ciRightBorder - ciLeftBorder ) < 0
* 
* @param ciLeftBorder the left border of the conf int.
* @param ciRightBorder the right border of the conf. int.
* @param initial_state the initial state for which the conf. int. was computed
* @param indiff_width the supremum of the allowed conf. int. width
* @param pIsIndiffErrStates the pointer to the boolean variable that should be TRUE if an error state exists
* @param pErrStatesBitset the bitset containing the indexes of error states
*/
inline void checkErrorConfIntState(const double ciLeftBorder, const double ciRightBorder, const int initial_state,
					const double indiff_width, BOOL * pIsIndiffErrStates, const bitset * pErrStatesBitset){
	if( ( ciRightBorder - ciLeftBorder ) < 0 || ( ( ciRightBorder - ciLeftBorder ) >= indiff_width ) ){
		*pIsIndiffErrStates = TRUE;
		set_bit_val(pErrStatesBitset, initial_state, BIT_ON);
	}
}

/****************************************************************************/
/**************THE INCREASE FUNCTION WITH THE STEP AND MAX BOUND*************/
/****************************************************************************/

/**
* This method does the following:
* 1. (* pTarget) += inc_step;
* 2. if (* pTarget) > max_val then
*	(* pTarget) = max_val;
* @param pTarget the pointer to the variable that will be increased.
* @param inc_step the increment step
* @param max_val the maximum value
*/
inline void increment( int * pTarget, const int inc_step, const int max_val ){
	IF_SAFETY( pTarget != NULL )
		(* pTarget) += inc_step;
		if( (* pTarget) > max_val ){
			(* pTarget) = max_val;
		}
	ELSE_SAFETY
		printf("ERROR: Trying to increment a value defined by a NULL pointer.\n");
		exit(1);
	ENDIF_SAFETY
}

/****************************************************************************/
/**************THE INCREASE FUNCTION WITH THE STEP AND MAX BOUND*************/
/****************************************************************************/

/**
* This method computes different conf intervals based on the provided values
* Note: this method allows to use two different parameters for the left and
* the right conf. int. borders. This is needed for computing the intervals
* in case of the unbounded-reachability problem.
* @param conf_zeta the zeta value derived from the confidence
* @param pCiLeftBorder the pointer to the left conf. int. border
* @param pCiRightBorder the pointer to the right conf. int. border
* @param left_param the parameter for computing the left conf. int. border
* @param right_param the parameter for computing the right conf. int. border
* @param theConfIntType indicate the kind of the conf. int. to be computed:
*				STANDARD_CONF_INT, AGRESTI_COULL_CONF_INT, etc.
*/
inline void computeBordersConfInt(const double conf_zeta, double * pCiLeftBorder,
					double * pCiRightBorder, const double sample_size,
					const double left_param, const double right_param,
					const int theConfIntType){
	double left_mean, right_mean, left_variance_sqrt, right_variance_sqrt;
	double work_sample_size = sample_size; /* For some conf. int.the sample size is changed */
	const double conf_zeta_squared = conf_zeta * conf_zeta;
	
	/* Compute the conf. int. borders */
	switch( theConfIntType ){
		case STANDARD_CONF_INT:
			left_mean = left_param / work_sample_size;
			right_mean = right_param / work_sample_size;
			left_variance_sqrt = sqrt( left_mean * ( sample_size - left_param ) / ( sample_size - 1.0 ) );
			right_variance_sqrt = sqrt( right_mean * ( sample_size - right_param ) / ( sample_size - 1.0 ) );
			break;
		case AGRESTI_COULL_CONF_INT:
			work_sample_size += conf_zeta_squared;
			left_mean = ( left_param + 0.5 * conf_zeta_squared ) / work_sample_size;
			right_mean = ( right_param + 0.5 * conf_zeta_squared ) / work_sample_size;
			left_variance_sqrt = sqrt( left_mean * ( 1.0 - left_mean ) );
			right_variance_sqrt = sqrt( right_mean * ( 1.0 - right_mean ) );
			break;
		default :
			printf("ERROR: An unexpected conf. int. type: %d.\n", theConfIntType);
			exit(1);
	}
	
	/* Compute the left border */
	( * pCiLeftBorder ) = left_mean - conf_zeta * left_variance_sqrt / sqrt(work_sample_size);
	/* Compute the right border */
	( * pCiRightBorder ) = right_mean + conf_zeta * right_variance_sqrt / sqrt(work_sample_size);
	
	/* Check if we've gotten values < 0.0 and > 1.0 */
	if( ( * pCiLeftBorder ) < 0.0 ){
		( * pCiLeftBorder ) = 0.0;
	}
	if( ( * pCiRightBorder ) > 1.0 ){
		( * pCiRightBorder ) = 1.0;
	}
}

/**
* This method computew the conf. int. borders. for the given sample, i.e.
* we compute and set: (*pCiLeftBorder), (*pCiRightBorder).
* NOTE: The borders that do not fit into the interval R[0,1] are truncated.
* @param conf_zeta the zeta value derived from the confidence
* @param pCiLeftBorder the pointer to the left conf. int. border
* @param pCiRightBorder the pointer to the right conf. int. border
* @param pSampleVecUntil the sample we use to derive the conf int
* @param theConfIntType indicate the kind of the conf. int. to be computed:
*				STANDARD_CONF_INT, AGRESTI_COULL_CONF_INT, etc.
*/
inline void computeBordersUU( const double conf_zeta, double * pCiLeftBorder, double * pCiRightBorder,
				const PTSampleVecUntil pSampleVecUntil, const int theConfIntType){
	IF_SAFETY( pSampleVecUntil != NULL )
		IF_SAFETY( ( pCiLeftBorder != NULL ) && ( pCiRightBorder != NULL ) )
			/* Compute the conf. int. for Bernoulli trials */
			computeBordersConfInt( conf_zeta, pCiLeftBorder, pCiRightBorder,
									(double) pSampleVecUntil->sampleVecBase.curr_sample_size,
									(double) pSampleVecUntil->sum_good,
									(double) ( pSampleVecUntil->sum_good + pSampleVecUntil->sum_trans ),
									theConfIntType );
		ELSE_SAFETY
			printf("ERROR: The left/right conf. int. border is passed by the NULL pointer.\n");
			exit(1);
		ENDIF_SAFETY
	ELSE_SAFETY
		printf("ERROR: The sample vector is passed by the NULL pointer.\n");
		exit(1);
	ENDIF_SAFETY
}

/**
* This method computes the conf. int. borders of steady-state simulation for 
* the given sample, i.e. we compute and set: (*pCiLeftBorder), (*pCiRightBorder).
* For more information see PhD Thesis of Ivan S. Zapreev.
* NOTE: The borders that do not fit into the interval R[0,1] are truncated.
* @param reg_cycles the number of samples, i.e. regeneration cycles calculated so far
* @param conf_zeta the zeta value derived from the confidence
* @param pCiLeftBorder the pointer to the left conf. int. border
* @param pCiRightBorder the pointer to the right conf. int. border
* @param pSampleVecSteady the sample we use to derive the conf int
*/
inline void computeBordersConfIntSS( const int reg_cycles, const double conf_zeta, double * pCiLeftBorder, 
					double * pCiRightBorder, const PTSampleVecSteady pSampleVecSteady ){
	int i;
	double point_estimate, half_width_ci; 
	double good_mean = 0; double time_mean = 0;
	double point_estimate_variance;
	double good_variance = 0; double time_variance = 0; double time_good_covariance = 0;
	double good_helper, time_helper;
	
	IF_SAFETY( pSampleVecSteady != NULL )
		IF_SAFETY( ( pCiLeftBorder != NULL ) && ( pCiRightBorder != NULL ) )
			
			/* Compute the mean value for "good" states and time */
			for( i = 0; i < reg_cycles; i++){
				good_mean += pSampleVecSteady->pGoodObservationsVec[i];
				time_mean += pSampleVecSteady->pTimeObservationsVec[i];
			}
			good_mean = good_mean / reg_cycles;
			time_mean = time_mean / reg_cycles;
			
			/* Compute the variances */
			for( i = 0; i < reg_cycles; i++){
				good_helper = pSampleVecSteady->pGoodObservationsVec[i] - good_mean;
				time_helper = pSampleVecSteady->pTimeObservationsVec[i] - time_mean;
				
				good_variance += pow( good_helper, 2);
				time_variance += pow( time_helper, 2);
				time_good_covariance += good_helper * time_helper; 
			}
			good_variance = good_variance / (reg_cycles - 1);
			time_variance = time_variance / (reg_cycles - 1);
			time_good_covariance = time_good_covariance / (reg_cycles - 1);
			
			/* Compute the point estimate */
			point_estimate = good_mean / time_mean;
			/* Compute the overall variance */
			point_estimate_variance = good_variance - 2 * point_estimate * time_good_covariance + pow(point_estimate, 2) * time_variance;
			
			/* Compute the half width of the conf. int. */
			half_width_ci = conf_zeta * sqrt(point_estimate_variance) / (time_mean * sqrt(reg_cycles));
			/* Compute the left border */
			( * pCiLeftBorder ) = point_estimate - half_width_ci;
			/* Compute the right border */
			( * pCiRightBorder ) = point_estimate + half_width_ci;
			
			/* Check if we've gotten values < 0.0 and > 1.0 */
			if( ( * pCiLeftBorder ) < 0.0 ){
				( * pCiLeftBorder ) = 0.0;
			}
			if( ( * pCiRightBorder ) > 1.0 ){
				( * pCiRightBorder ) = 1.0;
			}	
		ELSE_SAFETY
			printf("ERROR: The left/right conf. int. border is passed by the NULL pointer.\n");
			exit(1);
		ENDIF_SAFETY
	ELSE_SAFETY
		printf("ERROR: The sample vector is passed by the NULL pointer.\n");
		exit(1);
	ENDIF_SAFETY
}

/****************************************************************************/
/**********THE FUNCTION THAT SIMULATES THE SAMPLE FOR UNBOUNDED UNTIL********/
/****************************************************************************/
	
/* These are the same two values and they indicate that the simulation of */
/* new observations should start from the first one and also that all the */
/* observation in the sample were processed */
#define END_OF_OBSERVATIONS -1
#define FOR_ALL_OBSERVATIONS END_OF_OBSERVATIONS

/**
* This function is supposed to perform "depth_steps_needed" successive simulations
* for every observation after and including the "initial_obs_index"
* (if initial_obs_index != -1). The fields of "pSampleVecUntil", such as
*	sum_good, sum_trans, curr_simulation_depth, pTransStateInd, pObservationsVec.
* are updated correspondingly.
* WARNING: We simulate the states ASSUMING there are no self loops!
* @param pStateSpace the sparse matrix of the embedded DTMC with good ad bad
*			states made absorbing
* @param pSampleVecUntil the sample vector obtained on the previous iteration
* @param initial_obs_index the index of the observation we start simulating from
*				in the array pSampleVecUntil->pObservationsVec
* WARNING: For simulating all observation should be set to -1
* @param depth_steps_needed the simulation depth, i.e. the number of times we
*				simulate each observation in the array pObservationsVec
* @param pGoodStates the bitsets containing all the good absorbing states
* @param pTransientStates the bitsets containing all the transient states
*/
static void simulateUnbUntilSampleDTMC( const sparse * pStateSpace, PTSampleVecUntil pSampleVecUntil,
					const int initial_obs_index, const int depth_steps_needed,
					const bitset * pGoodStates, const bitset *pTransientStates ){
	/* Note that the "get_idx_next_non_zero" requires the previous found bitset element index */
	/* Therefore if we are not searching from the beginning of the bitset we have to use  */
	/* "initial_obs_index - 1" as the initial index */
	int curr_obs_idx = ( initial_obs_index == -1 ? initial_obs_index : initial_obs_index - 1 );
	const bitset * pTransStateInd = pSampleVecUntil->pTransStateInd;
	/* Will store the number of newly visited states */
	unsigned int newlyVisitedStates = 0;
	
	/* In case we are not adding new observations but just simulating all further */
	if( initial_obs_index == FOR_ALL_OBSERVATIONS ){
		/* Update the current simulation depth */
		pSampleVecUntil->curr_simulation_depth += depth_steps_needed;
	}
	
	/* For all the transient-state observation in the sample */
	while( ( curr_obs_idx = get_idx_next_non_zero( pTransStateInd, curr_obs_idx  )) != END_OF_OBSERVATIONS ){
		/* Take the current value of the curr_obs_idx's observation */
		int current_obs_state = pSampleVecUntil->pObservationsVec[curr_obs_idx];
		int extra_simulation_depth = 0;
		
		/* WARNING: DO NOT cast BITSET_BLOCK_TYPE into BOOL because get_bit_val */
		/* returns an "unsigned int" whereas BOOL is only "short"!!! */
		
		/* Iterate until we are in the absorbing state or we are on the right depth */
		while( ( extra_simulation_depth < depth_steps_needed ) && get_bit_val( pTransientStates, current_obs_state) ){
			/* Extract the corresponding row of the sparse matrix */
			values pOutgoingTrans = pStateSpace->val[ current_obs_state ];
			
			/* WARNING: We simulate the state ASSUMING there are no self loops! */
			/* WARNING: We simulate a pTransientStates state, i.e. there are outgoing */
			/* transitions in this state that are not self-loop transitions */
			current_obs_state = generateRandNumberDiscrete( pOutgoingTrans.col, pOutgoingTrans.val,
									pStateSpace->ncols[ current_obs_state ] );
			
			/* The next state was simulated */
			newlyVisitedStates++;
			
			/* Update the loop-condition variable */
			extra_simulation_depth += 1;
		}
		
		/* Update the observation pObservationsVec[curr_obs_idx] */
		pSampleVecUntil->pObservationsVec[curr_obs_idx] = current_obs_state;
		
		/* Update the sum_good, sum_trans and pTransStateInd fields */
		if( ! get_bit_val( pTransientStates, current_obs_state) ){
			pSampleVecUntil->sum_trans -= 1;
			set_bit_val( pTransStateInd, curr_obs_idx , FALSE);
			if( get_bit_val( pGoodStates, current_obs_state) ){
				pSampleVecUntil->sum_good += 1;
			}
		}
	}

	/* printSampleVectorUntil(pSampleVecUntil); */
	
	/* Update the number of states visited while preparing this sample */
	( (PTSampleVec) pSampleVecUntil)->num_visited_states += newlyVisitedStates;
}

/**
* This function is an implementation of the "extendSamples" function from the PhD Thesis
* of Ivan S. Zapreev. It either add new observations to the sample and simulates them
* or it increases the simulation depth of every observation in the sample.
* NOTE: This method can increase the sample size and the simulation depth at the same time.
* WARNING: We simulate the states ASSUMING there are no self loops!
* @param pStateSpace the sparse matrix of the embedded DTMC with good ad bad
*			states made absorbing
* @param pSampleVecUntil the sample vector obtained on the previous iteration
* @param new_sample_size the new sample size
* @param new_simulation_depth the new simulation depth
* @param pGoodStates the bitsets containing all the good absorbing states
* @param pTransientStates the bitsets containing all the transient states
*/
void simulateSampleVectorUnbUntilDTMC( const sparse * pStateSpace, PTSampleVecUntil pSampleVecUntil,
					const int new_sample_size, const int new_simulation_depth,
					const bitset * pGoodStates, const bitset * pTransientStates ){
	IF_SAFETY( pSampleVecUntil != NULL )
		/* Cast to the parent structure which contains initial_state and curr_sample_size */
		PTSampleVec pSampleVecUntilBase = (PTSampleVec) pSampleVecUntil;

		IF_SAFETY( ( pGoodStates != NULL ) && ( pTransientStates != NULL ) )
			IF_SAFETY( pStateSpace != NULL )
				int depth_steps_needed;

				/* If we need to extend the sample size */
				if( pSampleVecUntilBase->curr_sample_size < new_sample_size ){
					const int old_sample_size = pSampleVecUntilBase->curr_sample_size;
					
					/* Extend the sample size */
					extendSampleVectorUntil( pSampleVecUntil, new_sample_size );
					
					/* Do simulations for the newly added observations */
					simulateUnbUntilSampleDTMC( pStateSpace, pSampleVecUntil, old_sample_size,
									pSampleVecUntil->curr_simulation_depth,
									pGoodStates, pTransientStates );
				}
				
				/* If we need to increase the simulation depth */
				depth_steps_needed = new_simulation_depth - pSampleVecUntil->curr_simulation_depth;
				if( depth_steps_needed > 0 ){
					/* Do simulations for all the observations */
					simulateUnbUntilSampleDTMC( pStateSpace, pSampleVecUntil, FOR_ALL_OBSERVATIONS,
									depth_steps_needed, pGoodStates, pTransientStates );
				}
			ELSE_SAFETY
				printf("ERROR: The sparse matrix is NULL.\n");
				exit(1);
			ENDIF_SAFETY
		ELSE_SAFETY
			printf("ERROR: The good/transient states are undefined.\n");
			exit(1);
		ENDIF_SAFETY
	ELSE_SAFETY
		printf("ERROR: The sample that has to be simulated further is NULL.\n");
		exit(1);
	ENDIF_SAFETY
}
